@@ -16,7 +16,8 @@ import dataclasses as dc
 import dnf
 import datetime
 import tarfile
-
+import hashlib 
+import time
 from wheel_filename import parse_wheel_filename
 
 from .utils import *
@@ -36,9 +37,10 @@ class BinRegexps:
 
     def __post_init__(self):
         self.just_copy_re = []
-        for res_ in self.just_copy:
-            re_ = re.compile(res_ + '$')
-            self.just_copy_re.append(re_) 
+        if self.just_copy:
+            for res_ in self.just_copy:
+                re_ = re.compile(res_ + '$')
+                self.just_copy_re.append(re_) 
 
         self.need_patch_re = []
         for res_ in self.need_patch:
@@ -84,7 +86,7 @@ def fucking_magic(f):
 class PythonPackages:
     pip: list
     build: list
-    terra: list
+    terra: list = None
 
 @dc.dataclass
 class PackagesSpec:
@@ -249,6 +251,8 @@ class TerrariumAssembler:
         q_ = base.sack.query()
         self.installed_packages = q_.installed()
 
+        self.nuitka_plugins_dir = os.path.realpath(os.path.join(os.path.split(__file__)[0], '..', 'nuitka_plugins'))
+
         pass
 
 
@@ -278,6 +282,7 @@ class TerrariumAssembler:
 # Stage "%s"
 # Automatically called when terrarium_assembler --stage-%s "%s" 
 set -x
+sudo chmod a+rwx in/src  -R
 ''' % (desc, stage_, self.args.specfile))
             lf.write("\n".join(lines))
 
@@ -343,6 +348,7 @@ set -x
                 flags_ = target_.flags
             lines = []
             build_name = 'build_module_' + outputname
+            nuitka_plugins_dir = self.nuitka_plugins_dir
             lines.append("""
 export PATH="/usr/lib64/ccache:$PATH"
 find %(source_dir)s -name "*.py" | xargs -i{}  cksum {} > %(tmp_list)s
@@ -352,7 +358,7 @@ then
 """ % vars())
             lines.append(R"""
 else
-    python3 -m nuitka  %(nflags)s %(flags_)s  2>&1 >%(build_name)s.log
+    nice -19 python3 -m nuitka --include-plugin-directory=%(nuitka_plugins_dir)s %(nflags)s %(flags_)s  2>&1 >%(build_name)s.log
     RESULT=$?
     if [ $RESULT == 0 ]; then
         cp %(tmp_list)s %(target_list)s
@@ -380,7 +386,7 @@ export PATH="/usr/lib64/ccache:$PATH"
 """ % vars(self))
             build_name = 'build_' + srcname
             lines.append(R"""
-time python3 -m nuitka  %(nflags)s %(flags_)s %(src)s 2>&1 >%(build_name)s.log
+time nice -19 python3 -m nuitka  %(nflags)s %(flags_)s %(src)s 2>&1 >%(build_name)s.log
 """ % vars())
             self.fs.folders.append(target_dir)
             if "outputname" in target_:
@@ -481,6 +487,9 @@ rsync -rav  %(tmpdir_)s/modules/%(it)s/%(it)s.dist/ %(target_dir_)s/.
 
         if "__pycache__" in f:
             return False
+
+        if 'nginx' in f and 'sbin' in f:
+            w_ = 1
         
         if f == "": 
             return False
@@ -498,11 +507,13 @@ rsync -rav  %(tmpdir_)s/modules/%(it)s/%(it)s.dist/ %(target_dir_)s/.
     
         if len(parts) > 0 and parts[0] == "usr":
             parts.pop(0)
+            if len(parts) > 0 and parts[0] == "local":
+                parts.pop(0)
     
         if not parts:
             return False
     
-        if (parts[0] not in ["lib", "lib64"]) and (parts != ['bin', 'bash']):
+        if (parts[0] not in ["lib", "lib64"]) and (parts != ['bin', 'bash', 'sbin']):
             return False
         parts.pop(0)
     
@@ -514,11 +525,28 @@ rsync -rav  %(tmpdir_)s/modules/%(it)s/%(it)s.dist/ %(target_dir_)s/.
 
 
 
+    def rpm_update_time(self):
+        import time
+        for rpmdbpath in ["/var/lib/rpm/Packages", "/var/lib/rpm/rpmdb.sqlite"]:
+            if os.path.exists(rpmdbpath):
+                return str(time.ctime(os.path.getmtime(rpmdbpath)))        
+        return None        
+
+
     def dependencies(self, package_list, local=True):
         '''
         Генерируем список RPM-зависимостей для заданного списка пакетов.
         '''
+
         pl_ = self.packages2list(package_list)
+        package_list_md5 = hashlib.md5((self.rpm_update_time() + '\n' + '\n'.join(pl_)).encode('utf-8')).hexdigest()
+        cache_filename = 'cache_' + package_list_md5 + '.list' 
+        if os.path.exists(cache_filename):
+            with open(cache_filename, 'r', encoding='utf-8') as lf:
+                ls_ = lf.read()
+                list_ = ls_.split(',')
+                return list_
+
         repoch = re.compile("\d+\:")
         def remove_epoch(package):
             package_ = repoch.sub('', package)    
@@ -587,14 +615,47 @@ rsync -rav  %(tmpdir_)s/modules/%(it)s/%(it)s.dist/ %(target_dir_)s/.
 +-------------------------------+-----------------+""" % vars())            
                 pass
 
+        with open(cache_filename, 'w', encoding='utf-8') as lf:
+            lf.write(','.join(packages_))
 
         return packages_ 
 
-    def generate_file_list(self, packages):
+    def generate_file_list_from_pips(self, pips):
+        '''
+        Для заданного списка PIP-пакетов, возвращаем список файлов в этих пакетах, которые нужны нам.
+        '''
+        file_list = []
+        pips_ = [p.split('==')[0] for p in pips]
+        import pkg_resources
+        for dist in pkg_resources.working_set:
+            if dist.key in pips_:
+                if dist.has_metadata('RECORD'):
+                    lines = dist.get_metadata_lines('RECORD')
+                    paths = [line.split(',')[0] for line in lines]
+                    paths = [os.path.join(dist.location, p) for p in paths]
+                    file_list.extend(paths)
+
+        pass
+        res_ = [x for x in file_list if self.should_copy(x)]
+        return res_
+        pass
+
+
+    def generate_file_list_from_packages(self, packages):
         '''
         Для заданного списка RPM-файлов, возвращаем список файлов в этих пакетах, которые нужны нам.
         '''
     
+        package_list_md5 = hashlib.md5((self.rpm_update_time() + '\n' + '\n'.join(packages)).encode('utf-8')).hexdigest()
+        cache_filename = 'cachefilelist_' + package_list_md5 + '.list' 
+        if os.path.exists(cache_filename):
+            with open(cache_filename, 'r', encoding='utf-8') as lf:
+                ls_ = lf.read()
+                list_ = ls_.split('\n')
+                return list_
+
+
+
         exclusions = []
         for package_ in packages:
             exclusions += subprocess.check_output(['rpm', '-qd', package_], universal_newlines=True).splitlines()
@@ -606,6 +667,9 @@ rsync -rav  %(tmpdir_)s/modules/%(it)s/%(it)s.dist/ %(target_dir_)s/.
         # So we need a separate step in which all packages are added together.
     
         for package_ in packages:
+            if 'postgresql12-server' == package_:
+                wttt=1
+
             for try_ in range(3):
                 try:
                     files = subprocess.check_output(['repoquery',
@@ -632,6 +696,10 @@ rsync -rav  %(tmpdir_)s/modules/%(it)s/%(it)s.dist/ %(target_dir_)s/.
     
         pass
         res_ = [x for x in set(candidates) - set(exclusions) if self.should_copy(x)]
+
+        with open(cache_filename, 'w', encoding='utf-8') as lf:
+            lf.write('\n'.join(res_))
+
         return res_
 
     def add(self, what, to_=None, recursive=True):
@@ -715,19 +783,38 @@ rsync -rav  %(tmpdir_)s/modules/%(it)s/%(it)s.dist/ %(target_dir_)s/.
 
         args = self.args
         lines = []
+        lines2 = []
         in_src = os.path.relpath(self.src_dir, start=self.curdir)
         # lines.add("rm -rf %s " % in_src)
         lines.append("mkdir -p %s " % in_src)
         already_checkouted = set()
-        for td_ in self.pp.build + self.pp.terra + self.spec.templates_dirs:
+        for td_ in self.pp.build + (self.pp.terra if self.pp.terra else []) + self.spec.templates_dirs:
             git_url, git_branch, path_to_dir_, _ = self.explode_pp_node(td_)
             if path_to_dir_ not in already_checkouted:
+                probably_package_name = os.path.split(path_to_dir_)[-1]
                 already_checkouted.add(path_to_dir_)
                 path_to_dir = os.path.relpath(path_to_dir_, start=self.curdir)
                 newpath = path_to_dir + '.new'
                 lines.append('rm -rf "%(newpath)s"' % vars())
-                scmd = 'git --git-dir=/dev/null clone --single-branch --branch %(git_branch)s  --depth=1 %(git_url)s %(newpath)s ' % vars()
+                # scmd = 'git --git-dir=/dev/null clone --single-branch --branch %(git_branch)s  --depth=1 %(git_url)s %(newpath)s ' % vars()
+                scmd = '''
+git --git-dir=/dev/null clone  %(git_url)s %(newpath)s 
+pushd %(newpath)s 
+git checkout %(git_branch)s
+popd
+''' % vars()
                 lines.append(scmd)
+
+                lines2.append('''
+pushd "%(path_to_dir)s"
+git config core.fileMode false
+git pull
+sudo python -m pip uninstall  %(probably_package_name)s -y
+sudo python setup.py develop
+popd
+''' % vars())
+
+
                 # Fucking https://www.virtualbox.org/ticket/19086 + https://www.virtualbox.org/ticket/8761
                 lines.append("""
 if [ -d "%(newpath)s" ]; then
@@ -741,6 +828,7 @@ fi
 """ % vars())
 
         self.lines2sh("06-checkout", lines, 'checkout')    
+        self.lines2sh("96-pullall", lines2, 'checkout')    
         pass
 
     def explode_pp_node(self, td_):
@@ -781,25 +869,40 @@ fi
             pl_ = self.get_wheel_list_to_install()
             # pls_ = " ".join(pl_)
             for pls_ in pl_:
-                scmd = '%(root_dir)s/ebin/python3 -m pip install  %(pls_)s ' % vars()
+                if 'urllib3' in pls_:
+                    wt_ = 1
+                scmd = '%(root_dir)s/ebin/python3 -m pip install  %(pls_)s --force ' % vars()
+                print(scmd)
+                os.system(scmd)
+                wtf_path = f'{root_dir}/local/lib/python3.8/site-packages/enum'
+                if os.path.exists(wtf_path):
+                    print('Fucking enum34 here')
+                    sys.exit(0)
+
+        if self.pp.terra:
+            nodes_ = self.pp.terra
+            if self.args.debug:
+                nodes_ += self.pp.build
+            for td_ in nodes_:
+                git_url, git_branch, path_to_dir, setup_path = self.explode_pp_node(td_)
+        
+                os.chdir(setup_path)
+                # make_setup_if_not_exists()
+                release_mod = ''
+
+                #От отчаяния эвристика — пытаюсь выкинуть пакет перед инсталляцией из сорсов
+                # Вообще-то это не требуется и обычно работает без этого. Но иногда блядь, нет.
+                probably_package_name = os.path.split(setup_path)[-1]
+                scmd = "%(root_dir)s/ebin/python3 -m pip uninstall %(probably_package_name)s  -y " % vars()
                 print(scmd)
                 os.system(scmd)
 
-        nodes_ = self.pp.terra
-        if self.args.debug:
-            nodes_ += self.pp.build
-        for td_ in nodes_:
-            git_url, git_branch, path_to_dir, setup_path = self.explode_pp_node(td_)
-    
-            os.chdir(setup_path)
-            make_setup_if_not_exists()
-            release_mod = ''
-            # if self.args.release:
-            if not self.args.debug:
-                release_mod = ' --exclude-source-files '
-            scmd = "%(root_dir)s/ebin/python3 setup.py install --single-version-externally-managed  %(release_mod)s --root /   " % vars()
-            print(scmd)
-            os.system(scmd)
+                # if self.args.release:
+                if not self.args.debug:
+                    release_mod = ' --exclude-source-files '
+                scmd = "%(root_dir)s/ebin/python3 setup.py install --single-version-externally-managed  %(release_mod)s --root / --force   " % vars()
+                print(scmd)
+                os.system(scmd)
         pass
 
     def download_packages(self):
@@ -860,7 +963,7 @@ sudo dnf install --skip-broken %(in_bin)s/rpms/*.rpm -y --allowerasing
         wheelpath = os.path.join(self.in_bin, "ourwheel")
         relwheelpath = os.path.relpath(wheelpath, start=self.curdir)
         lines.append(R"rm -rf %(relwheelpath)s/*.*" % vars())
-        for td_, local_ in [ (x, True) for x in self.pp.build ] + [(x, False) for x in self.pp.terra]:
+        for td_, local_ in [ (x, True) for x in self.pp.build ] + [(x, False) for x in (self.pp.terra if self.pp.terra else [])]:
             git_url, git_branch, path_to_dir_, setup_path = self.explode_pp_node(td_)
             path_to_dir = os.path.relpath(path_to_dir_, start=self.curdir)
             relwheelpath = os.path.relpath(wheelpath, start=path_to_dir_)
@@ -904,13 +1007,16 @@ sudo dnf install --skip-broken %(in_bin)s/rpms/*.rpm -y --allowerasing
 
         pl_ = self.get_wheel_list_to_install()
         #--use-feature=2020-resolver
-        scmd = 'sudo python3 -m pip install --force-reinstall  %s ' % (" ".join(pl_))
+        scmd = 'sudo python3 -m pip install --no-dependencies --force-reinstall --ignore-installed  %s ' % (" ".join(pl_))
         lines.append(scmd)
 
         for p_ in pl_:
-            scmd = 'sudo python3 -m pip install --force-reinstall  %s ' % p_
+            scmd = 'sudo python3 -m pip install --no-dependencies --force-reinstall --ignore-installed  %s ' % p_
             lines.append(scmd)
 
+        scmd = 'sudo python3 -m pip uninstall -y enum34 ' 
+        lines.append(scmd)
+
         self.lines2sh("15-install-wheels", lines, "install-wheels")
         pass    
 
@@ -936,7 +1042,7 @@ rm -f %s/extwheel/*
             scmd = "python3 -m pip download %s --dest %s/extwheel " % (td_, bin_dir)
             lines.append(scmd)                
 
-        for td_, local_ in [ (x, True) for x in self.pp.build ] + [(x, False) for x in self.pp.terra]:
+        for td_, local_ in [ (x, True) for x in self.pp.build ] + [(x, False) for x in (self.pp.terra if self.pp.terra else []) ]:
             git_url, git_branch, path_to_dir, setup_path = self.explode_pp_node(td_)
             scmd = 'echo "\n\n** Downloading external wheels for %s" **\n' % path_to_dir
             lines.append(scmd)                
@@ -947,6 +1053,9 @@ rm -f %s/extwheel/*
                 lines.append(scmd)                
             pass
 
+        scmd = "rm -f %s/extwheel/enum34* " % (os.path.relpath(setup_path, start=self.curdir))
+        lines.append(scmd)                
+
         self.lines2sh("07-download-wheels", lines, "download-wheels")
         pass    
 
@@ -996,6 +1105,17 @@ rm -f %s/extwheel/*
         root_dir = self.root_dir 
         
         def install_templates(root_dir, args):
+            if 'copy_folders' in self.spec:
+                for it_ in self.spec.copy_folders:
+                    pass
+                from_ = os.path.join(self.src_dir, it_['from'])
+                to_ = os.path.join(root_dir, it_['to'])
+                mkdir_p(to_)
+
+                from distutils.dir_util import copy_tree
+                copy_tree(from_, to_)                        
+
+
             from jinja2 import Environment, FileSystemLoader, Template
             t.tic()
     
@@ -1071,15 +1191,29 @@ rm -f %s/extwheel/*
         specfile_ = self.args.specfile
         self.lines2sh("50-pack", [
             '''
+sudo chmod a+rx /usr/lib/cups -R           
 #terrarium_assembler --stage-pack=./out "%(specfile_)s" --stage-make-isoexe
 terrarium_assembler --stage-pack=./out "%(specfile_)s" 
             ''' % vars()])
 
+        self.lines2sh("51-pack-iso", [
+            '''
+sudo chmod a+rx /usr/lib/cups -R           
+terrarium_assembler --stage-pack=./out "%(specfile_)s" --stage-make-isoexe
+            ''' % vars()])
+
         self.lines2sh("91-pack-debug", [
             '''
+sudo chmod a+rx /usr/lib/cups -R           
 terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" 
             ''' % vars()])
 
+        self.lines2sh("92-pack-debug-iso", [
+            '''
+sudo chmod a+rx /usr/lib/cups -R           
+terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s" --stage-make-isoexe
+            ''' % vars()])
+
         root_dir = 'out'
         if self.args.stage_pack:
             root_dir = self.root_dir = expandpath(args.stage_pack)
@@ -1088,7 +1222,9 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s"
                 packages_to_deploy += self.ps.terra + self.ps.build
 
 
-            file_list = self.generate_file_list(self.dependencies(packages_to_deploy))
+            fs_ = self.generate_file_list_from_pips(self.pp.pip)
+            file_list = self.generate_file_list_from_packages(self.dependencies(packages_to_deploy))
+            file_list.extend(fs_)
 
             os.system('echo 2 > /proc/sys/vm/drop_caches ')
             if os.path.exists(root_dir + ".old"):
@@ -1101,9 +1237,12 @@ terrarium_assembler --debug --stage-pack=./out-debug "%(specfile_)s"
             mkdir_p(root_dir)    
         
             def copy_file_to_environment(f):
+                if '/usr/pgsql-12/bin/' in f:
+                    wtff = 1
+                if 'initdb' in f:
+                    wtff = 1
                 if not self.should_copy(f):
                     return
-                
                 if self.br.is_need_patch(f):  
                     self.process_binary(f)
                     self.add(f)
